@@ -21,11 +21,11 @@
 #include "ieee802_11.h"
 #include "wpa.h"
 #include "wme.h"
-#include "hostap_common.h"
 #include "beacon.h"
 #include "hw_features.h"
 #include "driver.h"
 #include "sta_info.h"
+#include "ieee802_11h.h"
 
 
 static u8 ieee802_11_erp_info(struct hostapd_data *hapd)
@@ -94,6 +94,74 @@ static u8 * hostapd_eid_erp_info(struct hostapd_data *hapd, u8 *eid)
 }
 
 
+static u8 * hostapd_eid_country(struct hostapd_data *hapd, u8 *eid,
+				int max_len)
+{
+	int left;
+	u8 *pos = eid;
+
+	if ((!hapd->iconf->ieee80211d && !hapd->iface->dfs_enable) ||
+	    max_len < 6)
+		return eid;
+
+	*pos++ = WLAN_EID_COUNTRY;
+	pos++; /* length will be set later */
+	memcpy(pos, hapd->iconf->country, 3); /* e.g., 'US ' */
+	pos += 3;
+	left = max_len - 3;
+
+	if ((pos - eid) & 1) {
+		if (left < 1)
+			return eid;
+		*pos++ = 0; /* pad for 16-bit alignment */
+		left--;
+	}
+
+	eid[1] = (pos - eid) - 2;
+
+	return pos;
+}
+
+
+static u8 * hostapd_eid_power_constraint(struct hostapd_data *hapd, u8 *eid)
+
+{
+	if (!hapd->iface->dfs_enable)
+		return eid;
+	*eid++ = WLAN_EID_PWR_CONSTRAINT;
+	*eid++ = 1;
+	*eid++ = hapd->iface->pwr_const;
+	return eid;
+}
+
+
+static u8 * hostapd_eid_tpc_report(struct hostapd_data *hapd, u8 *eid)
+
+{
+	if (!hapd->iface->dfs_enable)
+		return eid;
+	*eid++ = WLAN_EID_TPC_REPORT;
+	*eid++ = 2;
+	*eid++ = hapd->iface->tx_power; /* TX POWER */
+	*eid++ = 0; /* Link Margin */
+	return eid;
+}
+
+static u8 * hostapd_eid_channel_switch(struct hostapd_data *hapd, u8 *eid)
+
+{
+	if (!hapd->iface->dfs_enable || !hapd->iface->channel_switch)
+		return eid;
+	*eid++ = WLAN_EID_CHANNEL_SWITCH;
+	*eid++ = 3;
+	*eid++ = CHAN_SWITCH_MODE_QUIET;
+	*eid++ = hapd->iface->channel_switch; /* New channel */
+	/* 0 - very soon; 1 - before next TBTT; num - after num beacons */
+	*eid++ = 0;
+	return eid;
+}
+
+
 static u8 * hostapd_eid_wpa(struct hostapd_data *hapd, u8 *eid, size_t len,
 			    struct sta_info *sta)
 {
@@ -203,6 +271,11 @@ void handle_probe_req(struct hostapd_data *hapd, struct ieee80211_mgmt *mgmt,
 	/* DS Params */
 	pos = hostapd_eid_ds_params(hapd, pos);
 
+	pos = hostapd_eid_country(hapd, pos, epos - pos);
+
+	pos = hostapd_eid_power_constraint(hapd, pos);
+	pos = hostapd_eid_tpc_report(hapd, pos);
+
 	/* ERP Information element */
 	pos = hostapd_eid_erp_info(hapd, pos);
 
@@ -286,6 +359,13 @@ void ieee802_11_set_beacon(struct hostapd_data *hapd)
 
 	head_len = pos - (u8 *) head;
 
+	tailpos = hostapd_eid_country(hapd, tailpos,
+				      tail + BEACON_TAIL_BUF_SIZE - tailpos);
+
+	tailpos = hostapd_eid_power_constraint(hapd, tailpos);
+	tailpos = hostapd_eid_channel_switch(hapd, tailpos);
+	tailpos = hostapd_eid_tpc_report(hapd, tailpos);
+
 	/* ERP Information element */
 	tailpos = hostapd_eid_erp_info(hapd, tailpos);
 
@@ -319,7 +399,7 @@ void ieee802_11_set_beacon(struct hostapd_data *hapd)
 		printf("Failed to set Short Slot Time option in kernel "
 		       "driver\n");
 
-	if (hapd->iface->num_sta_no_short_preamble == 0 &&
+	if (hapd->iface && hapd->iface->num_sta_no_short_preamble == 0 &&
 	    hapd->iconf->preamble == SHORT_PREAMBLE)
 		preamble = SHORT_PREAMBLE;
 	else
@@ -331,7 +411,7 @@ void ieee802_11_set_beacon(struct hostapd_data *hapd)
 
 void ieee802_11_set_beacons(struct hostapd_iface *iface)
 {
-	int i;
+	size_t i;
 	for (i = 0; i < iface->num_bss; i++)
 		ieee802_11_set_beacon(iface->bss[i]);
 }
