@@ -28,6 +28,103 @@
 #define MAX_STA_COUNT 2007
 
 
+static int hostapd_config_read_vlan_file(struct hostapd_bss_config *bss,
+					 const char *fname)
+{
+	FILE *f;
+	char buf[128], *pos, *pos2;
+	int line = 0, vlan_id;
+	struct hostapd_vlan *vlan;
+
+	f = fopen(fname, "r");
+	if (!f) {
+		printf("VLAN file '%s' not readable.\n", fname);
+		return -1;
+	}
+
+	while (fgets(buf, sizeof(buf), f)) {
+		line++;
+
+		if (buf[0] == '#')
+			continue;
+		pos = buf;
+		while (*pos != '\0') {
+			if (*pos == '\n') {
+				*pos = '\0';
+				break;
+			}
+			pos++;
+		}
+		if (buf[0] == '\0')
+			continue;
+
+		if (buf[0] == '*') {
+			vlan_id = VLAN_ID_WILDCARD;
+			pos = buf + 1;
+		} else {
+			vlan_id = strtol(buf, &pos, 10);
+			if (buf == pos || vlan_id < 1 ||
+			    vlan_id > MAX_VLAN_ID) {
+				printf("Invalid VLAN ID at line %d in '%s'\n",
+				       line, fname);
+				fclose(f);
+				return -1;
+			}
+		}
+
+		while (*pos == ' ' || *pos == '\t')
+			pos++;
+		pos2 = pos;
+		while (*pos2 != ' ' && *pos2 != '\t' && *pos2 != '\0')
+			pos2++;
+		*pos2 = '\0';
+		if (*pos == '\0' || strlen(pos) > IFNAMSIZ) {
+			printf("Invalid VLAN ifname at line %d in '%s'\n",
+			       line, fname);
+			fclose(f);
+			return -1;
+		}
+
+		vlan = malloc(sizeof(*vlan));
+		if (vlan == NULL) {
+			printf("Out of memory while reading VLAN interfaces "
+			       "from '%s'\n", fname);
+			fclose(f);
+			return -1;
+		}
+
+		memset(vlan, 0, sizeof(*vlan));
+		vlan->vlan_id = vlan_id;
+		strncpy(vlan->ifname, pos, sizeof(vlan->ifname));
+		if (bss->vlan_tail)
+			bss->vlan_tail->next = vlan;
+		else
+			bss->vlan = vlan;
+		bss->vlan_tail = vlan;
+	}
+
+	fclose(f);
+
+	return 0;
+}
+
+
+static void hostapd_config_free_vlan(struct hostapd_bss_config *bss)
+{
+	struct hostapd_vlan *vlan, *prev;
+
+	vlan = bss->vlan;
+	prev = NULL;
+	while (vlan) {
+		prev = vlan;
+		vlan = vlan->next;
+		free(prev);
+	}
+
+	bss->vlan = NULL;
+}
+
+
 /* convert floats with one decimal place to value*10 int, i.e.,
  * "1.5" will return 15 */
 static int hostapd_config_read_int10(const char *value)
@@ -50,8 +147,6 @@ static int hostapd_config_read_int10(const char *value)
 
 static void hostapd_config_defaults_bss(struct hostapd_bss_config *bss)
 {
-	bss->radius = (struct hostapd_radius_servers *) (bss + 1);
-
 	bss->logger_syslog_level = HOSTAPD_LEVEL_INFO;
 	bss->logger_stdout_level = HOSTAPD_LEVEL_INFO;
 	bss->logger_syslog = (unsigned int) -1;
@@ -60,6 +155,9 @@ static void hostapd_config_defaults_bss(struct hostapd_bss_config *bss)
 	bss->auth_algs = HOSTAPD_AUTH_OPEN | HOSTAPD_AUTH_SHARED_KEY;
 
 	bss->wep_rekeying_period = 300;
+	/* use key0 in individual key and key1 in broadcast key */
+	bss->broadcast_key_idx_min = 1;
+	bss->broadcast_key_idx_max = 2;
 	bss->eap_reauth_period = 3600;
 
 	bss->wpa_group_rekey = 600;
@@ -70,6 +168,8 @@ static void hostapd_config_defaults_bss(struct hostapd_bss_config *bss)
 
 	bss->max_num_sta = MAX_STA_COUNT;
 
+	bss->dtim_period = 2;
+
 	bss->radius_server_auth_port = 1812;
 	bss->ap_max_inactivity = AP_MAX_INACTIVITY;
 	bss->eapol_version = EAPOL_VERSION;
@@ -92,7 +192,7 @@ static struct hostapd_config * hostapd_config_defaults(void)
 		{ aCWmin >> 2, aCWmin >> 1, 2, 1500 / 32, 1 };
 
 	conf = wpa_zalloc(sizeof(*conf));
-	bss = wpa_zalloc(sizeof(*bss) + sizeof(struct hostapd_radius_servers));
+	bss = wpa_zalloc(sizeof(*bss));
 	if (conf == NULL || bss == NULL) {
 		printf("Failed to allocate memory for configuration data.\n");
 		free(conf);
@@ -109,13 +209,25 @@ static struct hostapd_config * hostapd_config_defaults(void)
 		return NULL;
 	}
 
+	bss->radius = wpa_zalloc(sizeof(*bss->radius));
+	if (bss->radius == NULL) {
+		free(conf);
+		free(bss);
+		return NULL;
+	}
+
 	hostapd_config_defaults_bss(bss);
 
 	conf->num_bss = 1;
 	conf->bss = bss;
 
 	conf->beacon_int = 100;
+	conf->rts_threshold = -1; /* use driver default: 2347 */
+	conf->fragm_threshold = -1; /* user driver default: 2346 */
 	conf->send_probe_response = 1;
+	conf->bridge_packets = INTERNAL_BRIDGE_DO_NOT_CONTROL;
+
+	memcpy(conf->country, "US ", 3);
 
 	for (i = 0; i < NUM_TX_QUEUES; i++)
 		conf->tx_queue[i].aifs = -1; /* use hw default */
@@ -147,12 +259,19 @@ static int hostapd_parse_ip_addr(const char *txt, struct hostapd_ip_addr *addr)
 }
 
 
-static int mac_comp(const void *a, const void *b)
+int hostapd_mac_comp(const void *a, const void *b)
 {
 	return memcmp(a, b, sizeof(macaddr));
 }
 
 
+int hostapd_mac_comp_empty(const void *a)
+{
+	macaddr empty = { 0 };
+	return memcmp(a, empty, sizeof(macaddr));
+}
+
+
 static int hostapd_config_read_maclist(const char *fname, macaddr **acl,
 				       int *num)
 {
@@ -208,7 +327,7 @@ static int hostapd_config_read_maclist(const char *fname, macaddr **acl,
 
 	fclose(f);
 
-	qsort(*acl, *num, sizeof(macaddr), mac_comp);
+	qsort(*acl, *num, sizeof(macaddr), hostapd_mac_comp);
 
 	return 0;
 }
@@ -445,9 +564,9 @@ static int hostapd_config_read_eap_user(const char *fname,
 		}
 		num_methods = 0;
 		while (*start) {
-			char *pos2 = strchr(start, ',');
-			if (pos2) {
-				*pos2++ = '\0';
+			char *pos3 = strchr(start, ',');
+			if (pos3) {
+				*pos3++ = '\0';
 			}
 			user->methods[num_methods].method =
 				eap_get_type(start, &user->methods[num_methods]
@@ -464,9 +583,9 @@ static int hostapd_config_read_eap_user(const char *fname,
 			num_methods++;
 			if (num_methods >= EAP_USER_MAX_METHODS)
 				break;
-			if (pos2 == NULL)
+			if (pos3 == NULL)
 				break;
-			start = pos2;
+			start = pos3;
 		}
 		if (num_methods == 0) {
 			printf("No EAP types configured on line %d in '%s'\n",
@@ -724,23 +843,53 @@ static int hostapd_config_parse_cipher(int line, const char *value)
 }
 
 
-static int hostapd_config_check(struct hostapd_bss_config *conf)
+static int hostapd_config_check_bss(struct hostapd_bss_config *bss,
+				    struct hostapd_config *conf)
 {
-	if (conf->ieee802_1x && !conf->eap_server &&
-	    !conf->radius->auth_servers) {
+	if (bss->ieee802_1x && !bss->eap_server &&
+	    !bss->radius->auth_servers) {
 		printf("Invalid IEEE 802.1X configuration (no EAP "
 		       "authenticator configured).\n");
 		return -1;
 	}
 
-	if (conf->wpa && (conf->wpa_key_mgmt & WPA_KEY_MGMT_PSK) &&
-	    conf->ssid.wpa_psk == NULL && conf->ssid.wpa_passphrase == NULL &&
-	    conf->ssid.wpa_psk_file == NULL) {
+	if (bss->wpa && (bss->wpa_key_mgmt & WPA_KEY_MGMT_PSK) &&
+	    bss->ssid.wpa_psk == NULL && bss->ssid.wpa_passphrase == NULL &&
+	    bss->ssid.wpa_psk_file == NULL) {
 		printf("WPA-PSK enabled, but PSK or passphrase is not "
 		       "configured.\n");
 		return -1;
 	}
 
+	if (hostapd_mac_comp_empty(bss->bssid) != 0) {
+		size_t i;
+
+		for (i = 0; i < conf->num_bss; i++) {
+			if ((&conf->bss[i] != bss) &&
+			    (hostapd_mac_comp(conf->bss[i].bssid,
+					      bss->bssid) == 0)) {
+				printf("Duplicate BSSID " MACSTR
+				       " on interface '%s' and '%s'.\n",
+				       MAC2STR(bss->bssid),
+				       conf->bss[i].iface, bss->iface);
+				return -1;
+			}
+		}
+	}
+
+	return 0;
+}
+
+
+static int hostapd_config_check(struct hostapd_config *conf)
+{
+	size_t i;
+
+	for (i = 0; i < conf->num_bss; i++) {
+		if (hostapd_config_check_bss(&conf->bss[i], conf))
+			return -1;
+	}
+
 	return 0;
 }
 
@@ -819,6 +968,41 @@ static int hostapd_parse_rates(int **rate_list, char *val)
 }
 
 
+static int hostapd_config_bss(struct hostapd_config *conf, const char *ifname)
+{
+	struct hostapd_bss_config *bss;
+
+	if (*ifname == '\0')
+		return -1;
+
+	bss = realloc(conf->bss, (conf->num_bss + 1) *
+		      sizeof(struct hostapd_bss_config));
+	if (bss == NULL) {
+		printf("Failed to allocate memory for multi-BSS entry\n");
+		return -1;
+	}
+	conf->bss = bss;
+
+	bss = &(conf->bss[conf->num_bss]);
+	memset(bss, 0, sizeof(*bss));
+	bss->radius = wpa_zalloc(sizeof(*bss->radius));
+	if (bss->radius == NULL) {
+		printf("Failed to allocate memory for multi-BSS RADIUS "
+		       "data\n");
+		return -1;
+	}
+
+	conf->num_bss++;
+	conf->last_bss = bss;
+
+	hostapd_config_defaults_bss(bss);
+	snprintf(bss->iface, sizeof(bss->iface), "%s", ifname);
+	memcpy(bss->ssid.vlan, bss->iface, IFNAMSIZ + 1);
+
+	return 0;
+}
+
+
 static int valid_cw(int cw)
 {
 	return (cw == 1 || cw == 3 || cw == 7 || cw == 15 || cw == 31 ||
@@ -1041,15 +1225,17 @@ struct hostapd_config * hostapd_config_read(const char *fname)
 			bss->dump_log_name = strdup(pos);
 		} else if (strcmp(buf, "ssid") == 0) {
 			bss->ssid.ssid_len = strlen(pos);
-			if (bss->ssid.ssid_len >= HOSTAPD_MAX_SSID_LEN ||
+			if (bss->ssid.ssid_len > HOSTAPD_MAX_SSID_LEN ||
 			    bss->ssid.ssid_len < 1) {
 				printf("Line %d: invalid SSID '%s'\n", line,
 				       pos);
 				errors++;
+			} else {
+				memcpy(bss->ssid.ssid, pos,
+				       bss->ssid.ssid_len);
+				bss->ssid.ssid[bss->ssid.ssid_len] = '\0';
+				bss->ssid.ssid_set = 1;
 			}
-			memcpy(bss->ssid.ssid, pos, bss->ssid.ssid_len);
-			bss->ssid.ssid[bss->ssid.ssid_len] = '\0';
-			bss->ssid.ssid_set = 1;
 		} else if (strcmp(buf, "macaddr_acl") == 0) {
 			bss->macaddr_acl = atoi(pos);
 			if (bss->macaddr_acl != ACCEPT_UNLESS_DENIED &&
@@ -1078,6 +1264,14 @@ struct hostapd_config * hostapd_config_read(const char *fname)
 			}
 		} else if (strcmp(buf, "ap_max_inactivity") == 0) {
 			bss->ap_max_inactivity = atoi(pos);
+		} else if (strcmp(buf, "country_code") == 0) {
+			memcpy(conf->country, pos, 2);
+			/* FIX: make this configurable */
+			conf->country[2] = ' ';
+		} else if (strcmp(buf, "ieee80211d") == 0) {
+			conf->ieee80211d = atoi(pos);
+		} else if (strcmp(buf, "ieee80211h") == 0) {
+			conf->ieee80211h = atoi(pos);
 		} else if (strcmp(buf, "assoc_ap_addr") == 0) {
 			if (hwaddr_aton(pos, bss->assoc_ap_addr)) {
 				printf("Line %d: invalid MAC address '%s'\n",
@@ -1337,10 +1531,6 @@ struct hostapd_config * hostapd_config_read(const char *fname)
 		} else if (strcmp(buf, "rsn_preauth_interfaces") == 0) {
 			bss->rsn_preauth_interfaces = strdup(pos);
 #endif /* CONFIG_RSN_PREAUTH */
-#ifdef CONFIG_STAKEY
-		} else if (strcmp(buf, "stakey") == 0) {
-			bss->stakey = atoi(pos);
-#endif /* CONFIG_STAKEY */
 #ifdef CONFIG_PEERKEY
 		} else if (strcmp(buf, "peerkey") == 0) {
 			bss->peerkey = atoi(pos);
@@ -1418,6 +1608,29 @@ struct hostapd_config * hostapd_config_read(const char *fname)
 				errors++;
 			} else
 				conf->beacon_int = val;
+		} else if (strcmp(buf, "dtim_period") == 0) {
+			bss->dtim_period = atoi(pos);
+			if (bss->dtim_period < 1 || bss->dtim_period > 255) {
+				printf("Line %d: invalid dtim_period %d\n",
+				       line, bss->dtim_period);
+				errors++;
+			}
+		} else if (strcmp(buf, "rts_threshold") == 0) {
+			conf->rts_threshold = atoi(pos);
+			if (conf->rts_threshold < 0 ||
+			    conf->rts_threshold > 2347) {
+				printf("Line %d: invalid rts_threshold %d\n",
+				       line, conf->rts_threshold);
+				errors++;
+			}
+		} else if (strcmp(buf, "fragm_threshold") == 0) {
+			conf->fragm_threshold = atoi(pos);
+			if (conf->fragm_threshold < 256 ||
+			    conf->fragm_threshold > 2346) {
+				printf("Line %d: invalid fragm_threshold %d\n",
+				       line, conf->fragm_threshold);
+				errors++;
+			}
 		} else if (strcmp(buf, "send_probe_response") == 0) {
 			int val = atoi(pos);
 			if (val != 0 && val != 1) {
@@ -1437,6 +1650,8 @@ struct hostapd_config * hostapd_config_read(const char *fname)
 			}
 		} else if (strcmp(buf, "ignore_broadcast_ssid") == 0) {
 			bss->ignore_broadcast_ssid = atoi(pos);
+		} else if (strcmp(buf, "bridge_packets") == 0) {
+			conf->bridge_packets = atoi(pos);
 		} else if (strcmp(buf, "wep_default_key") == 0) {
 			bss->ssid.wep.idx = atoi(pos);
 			if (bss->ssid.wep.idx > 3) {
@@ -1456,6 +1671,16 @@ struct hostapd_config * hostapd_config_read(const char *fname)
 			}
 		} else if (strcmp(buf, "dynamic_vlan") == 0) {
 			bss->ssid.dynamic_vlan = atoi(pos);
+		} else if (strcmp(buf, "vlan_file") == 0) {
+			if (hostapd_config_read_vlan_file(bss, pos)) {
+				printf("Line %d: failed to read VLAN file "
+				       "'%s'\n", line, pos);
+				errors++;
+			}
+#ifdef CONFIG_FULL_DYNAMIC_VLAN
+		} else if (strcmp(buf, "vlan_tagged_interface") == 0) {
+			bss->ssid.vlan_tagged_interface = strdup(pos);
+#endif /* CONFIG_FULL_DYNAMIC_VLAN */
 		} else if (strcmp(buf, "passive_scan_interval") == 0) {
 			conf->passive_scan_interval = atoi(pos);
 		} else if (strcmp(buf, "passive_scan_listen") == 0) {
@@ -1480,6 +1705,24 @@ struct hostapd_config * hostapd_config_read(const char *fname)
 				       line);
 				errors++;
 			}
+		} else if (strcmp(buf, "bss") == 0) {
+			if (hostapd_config_bss(conf, pos)) {
+				printf("Line %d: invalid bss item\n", line);
+				errors++;
+			}
+		} else if (strcmp(buf, "bssid") == 0) {
+			if (bss == conf->bss) {
+				printf("Line %d: bssid item not allowed "
+				       "for the default interface\n", line);
+				errors++;
+			} else if (hwaddr_aton(pos, bss->bssid)) {
+				printf("Line %d: invalid bssid item\n", line);
+				errors++;
+			}
+#ifdef CONFIG_IEEE80211W
+		} else if (strcmp(buf, "ieee80211w") == 0) {
+			bss->ieee80211w = atoi(pos);
+#endif /* CONFIG_IEEE80211W */
 		} else {
 			printf("Line %d: unknown configuration item '%s'\n",
 			       line, buf);
@@ -1489,6 +1732,12 @@ struct hostapd_config * hostapd_config_read(const char *fname)
 
 	fclose(f);
 
+	if (bss->individual_wep_key_len == 0) {
+		/* individual keys are not use; can use key idx0 for broadcast
+		 * keys */
+		bss->broadcast_key_idx_min = 0;
+	}
+
 	for (i = 0; i < conf->num_bss; i++) {
 		bss = &conf->bss[i];
 
@@ -1506,11 +1755,11 @@ struct hostapd_config * hostapd_config_read(const char *fname)
 			bss->ssid.security_policy = SECURITY_STATIC_WEP;
 		else
 			bss->ssid.security_policy = SECURITY_PLAINTEXT;
-
-		if (hostapd_config_check(bss))
-			errors++;
 	}
 
+	if (hostapd_config_check(conf))
+		errors++;
+
 	if (errors) {
 		printf("%d errors found in configuration file '%s'\n",
 		       errors, fname);
@@ -1522,6 +1771,20 @@ struct hostapd_config * hostapd_config_read(const char *fname)
 }
 
 
+int hostapd_wep_key_cmp(struct hostapd_wep_keys *a, struct hostapd_wep_keys *b)
+{
+	int i;
+
+	if (a->idx != b->idx || a->default_len != b->default_len)
+		return 1;
+	for (i = 0; i < NUM_WEP_KEYS; i++)
+		if (a->len[i] != b->len[i] ||
+		    memcmp(a->key[i], b->key[i], a->len[i]) != 0)
+			return 1;
+	return 0;
+}
+
+
 static void hostapd_config_free_radius(struct hostapd_radius_server *servers,
 				       int num_servers)
 {
@@ -1542,6 +1805,16 @@ static void hostapd_config_free_eap_user(struct hostapd_eap_user *user)
 }
 
 
+static void hostapd_config_free_wep(struct hostapd_wep_keys *keys)
+{
+	int i;
+	for (i = 0; i < NUM_WEP_KEYS; i++) {
+		free(keys->key[i]);
+		keys->key[i] = NULL;
+	}
+}
+
+
 static void hostapd_config_free_bss(struct hostapd_bss_config *conf)
 {
 	struct hostapd_wpa_psk *psk, *prev;
@@ -1559,6 +1832,9 @@ static void hostapd_config_free_bss(struct hostapd_bss_config *conf)
 
 	free(conf->ssid.wpa_passphrase);
 	free(conf->ssid.wpa_psk_file);
+#ifdef CONFIG_FULL_DYNAMIC_VLAN
+	free(conf->ssid.vlan_tagged_interface);
+#endif /* CONFIG_FULL_DYNAMIC_VLAN */
 
 	user = conf->eap_user;
 	while (user) {
@@ -1585,6 +1861,20 @@ static void hostapd_config_free_bss(struct hostapd_bss_config *conf)
 	free(conf->eap_sim_db);
 	free(conf->radius_server_clients);
 	free(conf->test_socket);
+	free(conf->radius);
+	hostapd_config_free_vlan(conf);
+	if (conf->ssid.dyn_vlan_keys) {
+		struct hostapd_ssid *ssid = &conf->ssid;
+		size_t i;
+		for (i = 0; i <= ssid->max_dyn_vlan_keys; i++) {
+			if (ssid->dyn_vlan_keys[i] == NULL)
+				continue;
+			hostapd_config_free_wep(ssid->dyn_vlan_keys[i]);
+			free(ssid->dyn_vlan_keys[i]);
+		}
+		free(ssid->dyn_vlan_keys);
+		ssid->dyn_vlan_keys = NULL;
+	}
 }
 
 
@@ -1642,6 +1932,18 @@ int hostapd_rate_found(int *list, int rate)
 }
 
 
+const char * hostapd_get_vlan_id_ifname(struct hostapd_vlan *vlan, int vlan_id)
+{
+	struct hostapd_vlan *v = vlan;
+	while (v) {
+		if (v->vlan_id == vlan_id || v->vlan_id == VLAN_ID_WILDCARD)
+			return v->ifname;
+		v = v->next;
+	}
+	return NULL;
+}
+
+
 const u8 * hostapd_get_psk(const struct hostapd_bss_config *conf,
 			   const u8 *addr, const u8 *prev_psk)
 {
