@@ -2,7 +2,7 @@
  * hostapd / Driver interaction with MADWIFI 802.11 driver
  * Copyright (c) 2004, Sam Leffler <sam@errno.com>
  * Copyright (c) 2004, Video54 Technologies
- * Copyright (c) 2005, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2005-2006, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
@@ -45,7 +45,6 @@
 #include "priv_netlink.h"
 #include "sta_info.h"
 #include "l2_packet.h"
-#include "hostap_common.h"
 
 #include "eapol_sm.h"
 #include "wpa.h"
@@ -182,6 +181,8 @@ set80211param(struct madwifi_driver_data *drv, int op, int arg)
 
 	if (ioctl(drv->ioctl_sock, IEEE80211_IOCTL_SETPARAM, &iwr) < 0) {
 		perror("ioctl[IEEE80211_IOCTL_SETPARAM]");
+		wpa_printf(MSG_DEBUG, "%s: Failed to set parameter (op %d "
+			   "arg %d)", __func__, op, arg);
 		return -1;
 	}
 	return 0;
@@ -279,7 +280,7 @@ madwifi_configure_wpa(struct madwifi_driver_data *drv)
 	}
 
 	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
-		"%s: enable WPA= 0x%x\n", __func__, conf->wpa);
+		"%s: enable WPA=0x%x\n", __func__, conf->wpa);
 	if (set80211param(drv, IEEE80211_PARAM_WPA, conf->wpa)) {
 		printf("Unable to set WPA to %u\n", conf->wpa);
 		return -1;
@@ -292,11 +293,9 @@ static int
 madwifi_set_iface_flags(void *priv, int dev_up)
 {
 	struct madwifi_driver_data *drv = priv;
-	struct hostapd_data *hapd = drv->hapd;
 	struct ifreq ifr;
 
-	HOSTAPD_DEBUG(HOSTAPD_DEBUG_VERBOSE,
-		"%s: dev_up=%d\n", __func__, dev_up);
+	wpa_printf(MSG_DEBUG, "%s: dev_up=%d", __func__, dev_up);
 
 	if (drv->ioctl_sock < 0)
 		return -1;
@@ -364,11 +363,12 @@ madwifi_set_ieee8021x(const char *ifname, void *priv, int enabled)
 			HOSTAPD_LEVEL_WARNING, "Error enabling WPA/802.1X!");
 		return -1;
 	}
-	return madwifi_set_iface_flags(priv, 1);
+
+	return 0;
 }
 
 static int
-madwifi_set_privacy(void *priv, int enabled)
+madwifi_set_privacy(const char *ifname, void *priv, int enabled)
 {
 	struct madwifi_driver_data *drv = priv;
 	struct hostapd_data *hapd = drv->hapd;
@@ -385,6 +385,7 @@ madwifi_set_sta_authorized(void *priv, const u8 *addr, int authorized)
 	struct madwifi_driver_data *drv = priv;
 	struct hostapd_data *hapd = drv->hapd;
 	struct ieee80211req_mlme mlme;
+	int ret;
 
 	HOSTAPD_DEBUG(HOSTAPD_DEBUG_VERBOSE,
 		"%s: addr=%s authorized=%d\n",
@@ -396,8 +397,14 @@ madwifi_set_sta_authorized(void *priv, const u8 *addr, int authorized)
 		mlme.im_op = IEEE80211_MLME_UNAUTHORIZE;
 	mlme.im_reason = 0;
 	memcpy(mlme.im_macaddr, addr, IEEE80211_ADDR_LEN);
-	return set80211priv(priv, IEEE80211_IOCTL_SETMLME, &mlme,
-			    sizeof(mlme));
+	ret = set80211priv(priv, IEEE80211_IOCTL_SETMLME, &mlme,
+			   sizeof(mlme));
+	if (ret < 0) {
+		wpa_printf(MSG_DEBUG, "%s: Failed to %sauthorize STA " MACSTR,
+			   __func__, authorized ? "" : "un", MAC2STR(addr));
+	}
+
+	return ret;
 }
 
 static int
@@ -417,6 +424,7 @@ madwifi_del_key(void *priv, const u8 *addr, int key_idx)
 	struct madwifi_driver_data *drv = priv;
 	struct hostapd_data *hapd = drv->hapd;
 	struct ieee80211req_del_key wk;
+	int ret;
 
 	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
 		"%s: addr=%s key_idx=%d\n",
@@ -430,7 +438,13 @@ madwifi_del_key(void *priv, const u8 *addr, int key_idx)
 		wk.idk_keyix = key_idx;
 	}
 
-	return set80211priv(priv, IEEE80211_IOCTL_DELKEY, &wk, sizeof(wk));
+	ret = set80211priv(priv, IEEE80211_IOCTL_DELKEY, &wk, sizeof(wk));
+	if (ret < 0) {
+		wpa_printf(MSG_DEBUG, "%s: Failed to delete key (addr " MACSTR
+			   " key_idx %d)", __func__, MAC2STR(addr), key_idx);
+	}
+
+	return ret;
 }
 
 static int
@@ -442,6 +456,7 @@ madwifi_set_key(const char *ifname, void *priv, const char *alg,
 	struct hostapd_data *hapd = drv->hapd;
 	struct ieee80211req_key wk;
 	u_int8_t cipher;
+	int ret;
 
 	if (strcmp(alg, "none") == 0)
 		return madwifi_del_key(priv, addr, key_idx);
@@ -482,7 +497,15 @@ madwifi_set_key(const char *ifname, void *priv, const char *alg,
 	wk.ik_keylen = key_len;
 	memcpy(wk.ik_keydata, key, key_len);
 
-	return set80211priv(priv, IEEE80211_IOCTL_SETKEY, &wk, sizeof(wk));
+	ret = set80211priv(priv, IEEE80211_IOCTL_SETKEY, &wk, sizeof(wk));
+	if (ret < 0) {
+		wpa_printf(MSG_DEBUG, "%s: Failed to set key (addr " MACSTR
+			   " key_idx %d alg '%s' key_len %lu txkey %d)",
+			   __func__, MAC2STR(wk.ik_macaddr), key_idx, alg,
+			   (unsigned long) key_len, txkey);
+	}
+
+	return ret;
 }
 
 
@@ -505,7 +528,9 @@ madwifi_get_seqnum(const char *ifname, void *priv, const u8 *addr, int idx,
 	wk.ik_keyix = idx;
 
 	if (set80211priv(priv, IEEE80211_IOCTL_GETKEY, &wk, sizeof(wk))) {
-		printf("Failed to get encryption.\n");
+		wpa_printf(MSG_DEBUG, "%s: Failed to get encryption data "
+			   "(addr " MACSTR " key_idx %d)",
+			   __func__, MAC2STR(wk.ik_macaddr), idx);
 		return -1;
 	}
 
@@ -565,6 +590,8 @@ madwifi_read_sta_driver_data(void *priv, struct hostap_sta_driver_data *data,
 			 IEEE80211_IOCTL_GETSTASTATS,
 #endif /* MADWIFI_NG */
 			 &stats, sizeof(stats))) {
+		wpa_printf(MSG_DEBUG, "%s: Failed to fetch STA stats (addr "
+			   MACSTR ")", __func__, MAC2STR(addr));
 		if (memcmp(addr, drv->acct_mac, ETH_ALEN) == 0) {
 			memcpy(data, &drv->acct_data, sizeof(*data));
 			return 0;
@@ -631,14 +658,21 @@ madwifi_sta_clear_stats(void *priv, const u8 *addr)
 	struct madwifi_driver_data *drv = priv;
 	struct hostapd_data *hapd = drv->hapd;
 	struct ieee80211req_mlme mlme;
-	
+	int ret;
+
 	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL, "%s: addr=%s\n",
 		      __func__, ether_sprintf(addr));
 
 	mlme.im_op = IEEE80211_MLME_CLEAR_STATS;
 	memcpy(mlme.im_macaddr, addr, IEEE80211_ADDR_LEN);
-	return set80211priv(priv, IEEE80211_IOCTL_SETMLME, &mlme,
-			    sizeof(mlme));
+	ret = set80211priv(priv, IEEE80211_IOCTL_SETMLME, &mlme,
+			   sizeof(mlme));
+	if (ret < 0) {
+		wpa_printf(MSG_DEBUG, "%s: Failed to clear STA stats (addr "
+			   MACSTR ")", __func__, MAC2STR(addr));
+	}
+
+	return ret;
 #else /* MADWIFI_BSD && IEEE80211_MLME_CLEAR_STATS */
 	return 0; /* FIX */
 #endif /* MADWIFI_BSD && IEEE80211_MLME_CLEAR_STATS */
@@ -646,7 +680,7 @@ madwifi_sta_clear_stats(void *priv, const u8 *addr)
 
 
 static int
-madwifi_set_opt_ie(void *priv, const u8 *ie, size_t ie_len)
+madwifi_set_opt_ie(const char *ifname, void *priv, const u8 *ie, size_t ie_len)
 {
 	/*
 	 * Do nothing; we setup parameters at startup that define the
@@ -661,6 +695,7 @@ madwifi_sta_deauth(void *priv, const u8 *addr, int reason_code)
 	struct madwifi_driver_data *drv = priv;
 	struct hostapd_data *hapd = drv->hapd;
 	struct ieee80211req_mlme mlme;
+	int ret;
 
 	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
 		"%s: addr=%s reason_code=%d\n",
@@ -669,7 +704,14 @@ madwifi_sta_deauth(void *priv, const u8 *addr, int reason_code)
 	mlme.im_op = IEEE80211_MLME_DEAUTH;
 	mlme.im_reason = reason_code;
 	memcpy(mlme.im_macaddr, addr, IEEE80211_ADDR_LEN);
-	return set80211priv(priv, IEEE80211_IOCTL_SETMLME, &mlme, sizeof(mlme));
+	ret = set80211priv(priv, IEEE80211_IOCTL_SETMLME, &mlme, sizeof(mlme));
+	if (ret < 0) {
+		wpa_printf(MSG_DEBUG, "%s: Failed to deauth STA (addr " MACSTR
+			   " reason %d)",
+			   __func__, MAC2STR(addr), reason_code);
+	}
+
+	return ret;
 }
 
 static int
@@ -678,14 +720,23 @@ madwifi_sta_disassoc(void *priv, const u8 *addr, int reason_code)
 	struct madwifi_driver_data *drv = priv;
 	struct hostapd_data *hapd = drv->hapd;
 	struct ieee80211req_mlme mlme;
+	int ret;
 
 	HOSTAPD_DEBUG(HOSTAPD_DEBUG_MINIMAL,
 		"%s: addr=%s reason_code=%d\n",
 		__func__, ether_sprintf(addr), reason_code);
 
+	mlme.im_op = IEEE80211_MLME_DISASSOC;
 	mlme.im_reason = reason_code;
 	memcpy(mlme.im_macaddr, addr, IEEE80211_ADDR_LEN);
-	return set80211priv(priv, IEEE80211_IOCTL_SETMLME, &mlme, sizeof(mlme));
+	ret = set80211priv(priv, IEEE80211_IOCTL_SETMLME, &mlme, sizeof(mlme));
+	if (ret < 0) {
+		wpa_printf(MSG_DEBUG, "%s: Failed to disassoc STA (addr "
+			   MACSTR " reason %d)",
+			   __func__, MAC2STR(addr), reason_code);
+	}
+
+	return ret;
 }
 
 static int
@@ -699,7 +750,7 @@ madwifi_del_sta(struct madwifi_driver_data *drv, u8 addr[IEEE80211_ADDR_LEN])
 
 	sta = ap_get_sta(hapd, addr);
 	if (sta != NULL) {
-		sta->flags &= ~WLAN_STA_ASSOC;
+		sta->flags &= ~(WLAN_STA_AUTH | WLAN_STA_ASSOC);
 		wpa_auth_sm_event(sta->wpa_sm, WPA_DISASSOC);
 		sta->acct_terminate_cause = RADIUS_ACCT_TERMINATE_CAUSE_USER_REQUEST;
 		ieee802_1x_notify_port_enabled(sta->eapol_sm, 0);
@@ -722,6 +773,8 @@ madwifi_process_wpa_ie(struct madwifi_driver_data *drv, struct sta_info *sta)
 	memset(&ie, 0, sizeof(ie));
 	memcpy(ie.wpa_macaddr, sta->addr, IEEE80211_ADDR_LEN);
 	if (set80211priv(drv, IEEE80211_IOCTL_GETWPAIE, &ie, sizeof(ie))) {
+		wpa_printf(MSG_ERROR, "%s: Failed to get WPA/RSN IE",
+			   __func__);
 		printf("Failed to get WPA/RSN information element.\n");
 		return -1;		/* XXX not right */
 	}
@@ -790,7 +843,7 @@ madwifi_new_sta(struct madwifi_driver_data *drv, u8 addr[IEEE80211_ADDR_LEN])
 	 * kick the authenticator into action.
 	 */
 	new_assoc = (sta->flags & WLAN_STA_ASSOC) == 0;
-	sta->flags |= WLAN_STA_ASSOC;
+	sta->flags |= WLAN_STA_AUTH | WLAN_STA_ASSOC;
 	wpa_auth_sm_event(sta->wpa_sm, WPA_ASSOC);
 	hostapd_new_assoc_sta(hapd, sta, !new_assoc);
 	ieee802_1x_notify_port_enabled(sta->eapol_sm, 1);
@@ -1089,7 +1142,7 @@ madwifi_wireless_event_deinit(void *priv)
 
 static int
 madwifi_send_eapol(void *priv, const u8 *addr, const u8 *data, size_t data_len,
-		   int encrypt)
+		   int encrypt, const u8 *own_addr)
 {
 	struct madwifi_driver_data *drv = priv;
 	unsigned char buf[3000];
@@ -1115,7 +1168,7 @@ madwifi_send_eapol(void *priv, const u8 *addr, const u8 *data, size_t data_len,
 	}
 	eth = (struct l2_ethhdr *) bp;
 	memcpy(eth->h_dest, addr, ETH_ALEN);
-	memcpy(eth->h_source, drv->hapd->own_addr, ETH_ALEN);
+	memcpy(eth->h_source, own_addr, ETH_ALEN);
 	eth->h_proto = htons(ETH_P_EAPOL);
 	memcpy(eth+1, data, data_len);
 
@@ -1206,6 +1259,7 @@ madwifi_init(struct hostapd_data *hapd)
 	}
 
 	madwifi_set_iface_flags(drv, 0);	/* mark down during setup */
+	madwifi_set_privacy(drv->iface, drv, 0); /* default to no privacy */
 
 	hapd->driver = &drv->ops;
 	return 0;
@@ -1238,7 +1292,7 @@ madwifi_deinit(void *priv)
 }
 
 static int
-madwifi_set_ssid(void *priv, const u8 *buf, int len)
+madwifi_set_ssid(const char *ifname, void *priv, const u8 *buf, int len)
 {
 	struct madwifi_driver_data *drv = priv;
 	struct iwreq iwr;
@@ -1258,7 +1312,7 @@ madwifi_set_ssid(void *priv, const u8 *buf, int len)
 }
 
 static int
-madwifi_get_ssid(void *priv, u8 *buf, int len)
+madwifi_get_ssid(const char *ifname, void *priv, u8 *buf, int len)
 {
 	struct madwifi_driver_data *drv = priv;
 	struct iwreq iwr;
@@ -1286,6 +1340,12 @@ madwifi_set_countermeasures(void *priv, int enabled)
 	return set80211param(drv, IEEE80211_PARAM_COUNTERMEASURES, enabled);
 }
 
+static int
+madwifi_commit(void *priv)
+{
+	return madwifi_set_iface_flags(priv, 1);
+}
+
 static const struct driver_ops madwifi_driver_ops = {
 	.name			= "madwifi",
 	.init			= madwifi_init,
@@ -1307,6 +1367,7 @@ static const struct driver_ops madwifi_driver_ops = {
 	.get_ssid		= madwifi_get_ssid,
 	.set_countermeasures	= madwifi_set_countermeasures,
 	.sta_clear_stats        = madwifi_sta_clear_stats,
+	.commit			= madwifi_commit,
 };
 
 void madwifi_driver_register(void)
