@@ -41,7 +41,6 @@
 #include "radius.h"
 #include "sta_info.h"
 #include "accounting.h"
-#include "hostap_common.h"
 
 const int PIM_BUF_SIZE = 4096;
 
@@ -114,6 +113,7 @@ static int prism54_waitpim(void *priv, unsigned long oid, void *buf, int len,
 						return -1;
 					}
 				} else {
+					gettimeofday(&ctv, NULL);
 					continue;
 				}
 			}
@@ -128,7 +128,8 @@ static int prism54_waitpim(void *priv, unsigned long oid, void *buf, int len,
 
 /* send an eapol packet */
 static int prism54_send_eapol(void *priv, const u8 *addr,
-			      const u8 *data, size_t data_len, int encrypt)
+			      const u8 *data, size_t data_len, int encrypt,
+			      const u8 *own_addr)
 {
 	struct prism54_driver_data *drv = priv;
 	ieee802_3_hdr *hdr;
@@ -145,7 +146,7 @@ static int prism54_send_eapol(void *priv, const u8 *addr,
 	}
 
 	memcpy(&hdr->da[0], addr, ETH_ALEN);
-	memcpy(&hdr->sa[0], drv->hapd->own_addr, ETH_ALEN);
+	memcpy(&hdr->sa[0], own_addr, ETH_ALEN);
 	hdr->type = htons(ETH_P_PAE);
 	pos = (u8 *) (hdr + 1);
 	memcpy(pos, data, data_len);
@@ -234,6 +235,7 @@ static int prism54_set_encryption(const char *ifname, void *priv,
 		/* the only way to clear the key is to deauth it */
 		/* and prism54 is capable to receive unencrypted packet */
 		/* so we do nothing here */
+		free(hdr);
 		return 0;
 	} else {
 		printf("bad auth type: %s\n", alg);
@@ -250,8 +252,10 @@ static int prism54_set_encryption(const char *ifname, void *priv,
 	memcpy(buf, key, key_len);
 	
 	ret = send(drv->pim_sock, hdr, blen, 0);
-	if (ret < 0)
+	if (ret < 0) {
+		free(hdr);
 		return ret;
+	}
 	prism54_waitpim(priv, hdr->oid, hdr, blen, 10);
 
 	free(hdr);
@@ -344,7 +348,8 @@ static int prism54_init_1x(void *priv)
 }
 
 
-static int prism54_set_privacy_invoked(void *priv, int flag)
+static int prism54_set_privacy_invoked(const char *ifname, void *priv,
+				       int flag)
 {
 	struct prism54_driver_data *drv = priv;
 	pimdev_hdr *hdr;
@@ -372,7 +377,8 @@ static int prism54_set_privacy_invoked(void *priv, int flag)
 }
 
  
-static int prism54_ioctl_setiwessid(void *priv, const u8 *buf, int len)
+static int prism54_ioctl_setiwessid(const char *ifname, void *priv,
+				    const u8 *buf, int len)
 {
 #if 0
 	struct prism54_driver_data *drv = priv;
@@ -404,8 +410,9 @@ static int prism54_flush(void *priv)
 	int ret, i;
 	long *nsta;
 	int blen = sizeof(*hdr) + sizeof(*mlme);
+	char *mac_id;
 
-	hdr = malloc(blen);
+	hdr = wpa_zalloc(blen);
 	if (hdr == NULL)
 		return -1;
 
@@ -422,7 +429,9 @@ static int prism54_flush(void *priv)
 	}
 	for (i = 0; i < le_to_host32(*nsta); i++) {
 		mlme->id = -1;
-		memcpy(&mlme->address[0], mac_id_get(drv, i), ETH_ALEN);
+		mac_id = mac_id_get(drv, i);
+		if (mac_id)
+			memcpy(&mlme->address[0], mac_id, ETH_ALEN);
 		mlme->code = host_to_le16(WLAN_REASON_UNSPECIFIED);
 		mlme->state = htons(DOT11_STATE_NONE);
 		mlme->size = 0;
@@ -434,7 +443,9 @@ static int prism54_flush(void *priv)
 	}
 	for (i = 0; i < le_to_host32(*nsta); i++) {
 		mlme->id = -1;
-		memcpy(&mlme->address[0], mac_id_get(drv, i), ETH_ALEN);
+		mac_id = mac_id_get(drv, i);
+		if (mac_id)
+			memcpy(&mlme->address[0], mac_id, ETH_ALEN);
 		mlme->code = host_to_le16(WLAN_REASON_UNSPECIFIED);
 		mlme->state = htons(DOT11_STATE_NONE);
 		mlme->size = 0;
@@ -532,7 +543,7 @@ static int prism54_get_inact_sec(void *priv, const u8 *addr)
 
 
 /* set attachments */
-static int prism54_set_generic_elem(void *priv,
+static int prism54_set_generic_elem(const char *ifname, void *priv,
 				    const u8 *elem, size_t elem_len)
 {
 	struct prism54_driver_data *drv = priv;
@@ -596,13 +607,15 @@ static void prism54_handle_deauth(struct prism54_driver_data *drv,
 	struct obj_mlme *mlme;
 	pimdev_hdr *hdr;
 	struct sta_info *sta;
+	char *mac_id;
+
 	hdr = (pimdev_hdr *) buf;
 	mlme = (struct obj_mlme *) &hdr[1];
 	sta = ap_get_sta(drv->hapd, (u8 *) &mlme->address[0]);
-	memcpy(&mlme->address[0], mac_id_get(drv, mlme->id), ETH_ALEN);
-	if (sta == NULL) {
+	mac_id = mac_id_get(drv, mlme->id);
+	if (sta == NULL || mac_id == NULL)
 		return;
-	}
+	memcpy(&mlme->address[0], mac_id, ETH_ALEN);
 	sta->flags &= ~(WLAN_STA_AUTH | WLAN_STA_ASSOC);
 	wpa_auth_sm_event(sta->wpa_sm, WPA_DEAUTH);
 	sta->acct_terminate_cause = RADIUS_ACCT_TERMINATE_CAUSE_USER_REQUEST;
@@ -617,9 +630,14 @@ static void prism54_handle_disassoc(struct prism54_driver_data *drv,
 	struct obj_mlme *mlme;
 	pimdev_hdr *hdr;
 	struct sta_info *sta;
+	char *mac_id;
+
 	hdr = (pimdev_hdr *) buf;
 	mlme = (struct obj_mlme *) &hdr[1];
-	memcpy(&mlme->address[0], mac_id_get(drv, mlme->id), ETH_ALEN);
+	mac_id = mac_id_get(drv, mlme->id);
+	if (mac_id == NULL)
+		return;
+	memcpy(&mlme->address[0], mac_id, ETH_ALEN);
 	sta = ap_get_sta(drv->hapd, (u8 *) &mlme->address[0]);
 	if (sta == NULL) {
 		return;
@@ -679,7 +697,8 @@ fail:
 	printf("auth fail: %x\n", resp);
 	mlme->code = host_to_le16(resp);
 	mlme->size = 0;
-	sta->flags &= ~(WLAN_STA_AUTH | WLAN_STA_ASSOC);
+	if (sta)
+		sta->flags &= ~(WLAN_STA_AUTH | WLAN_STA_ASSOC);
 	hdr->oid = htonl(DOT11_OID_DEAUTHENTICATEEX);
 	hdr->op = htonl(PIMOP_SET);
 	send(drv->pim_sock, hdr, sizeof(*hdr)+sizeof(*mlme), 0);
@@ -699,6 +718,8 @@ static void prism54_handle_assoc(struct prism54_driver_data *drv,
 	int ieofs = 0;
 	size_t wpa_ie_len;
 	int resp, new_assoc;
+	char *mac_id;
+
 	resp = 0;
 	hdr = (pimdev_hdr *) buf;
 	mlme = (struct obj_mlmeex *) &hdr[1];
@@ -715,7 +736,10 @@ static void prism54_handle_assoc(struct prism54_driver_data *drv,
 			printf("bad assoc packet\n");
 			return;
 		}
-		memcpy(&mlme->address[0], mac_id_get(drv, mlme->id), ETH_ALEN);
+		mac_id = mac_id_get(drv, mlme->id);
+		if (mac_id == NULL)
+			return;
+		memcpy(&mlme->address[0], mac_id, ETH_ALEN);
 		sta = ap_get_sta(drv->hapd, (u8 *) &mlme->address[0]);
 		if (sta == NULL) {
 			printf("cannot get sta\n");
@@ -801,14 +825,17 @@ static void prism54_handle_assoc(struct prism54_driver_data *drv,
 			printf("bad assoc packet\n");
 			return;
 		}
-		memcpy(&mlme->address[0], mac_id_get(drv, mlme->id), ETH_ALEN);
+		mac_id = mac_id_get(drv, mlme->id);
+		if (mac_id == NULL)
+			return;
+		memcpy(&mlme->address[0], mac_id, ETH_ALEN);
 		sta = ap_get_sta(drv->hapd, (u8 *) &mlme->address[0]);
 		if (sta == NULL) {
 			printf("cannot get sta\n");
 			return;
 		}
 		new_assoc = (sta->flags & WLAN_STA_ASSOC) == 0;
-		sta->flags |= WLAN_STA_ASSOC;
+		sta->flags |= WLAN_STA_AUTH | WLAN_STA_ASSOC;
 		wpa_auth_sm_event(sta->wpa_sm, WPA_ASSOC);
 		hostapd_new_assoc_sta(drv->hapd, sta, !new_assoc);
 		ieee802_1x_notify_port_enabled(sta->eapol_sm, 1);
