@@ -233,6 +233,13 @@ int eap_server_register_methods(void)
 	}
 #endif /* EAP_SAKE */
 
+#ifdef EAP_GPSK
+	if (ret == 0) {
+		int eap_server_gpsk_register(void);
+		ret = eap_server_gpsk_register();
+	}
+#endif /* EAP_GPSK */
+
 #ifdef EAP_VENDOR_TEST
 	if (ret == 0) {
 		int eap_server_vendor_test_register(void);
