@@ -55,35 +55,23 @@
 #include "common.h"
 #include "milenage.h"
 
-/* Define AKE_USE_MILENAGE to enable Milenage authentication algorithms for
- * testing real USIM cards. */
-/* #define AKA_USE_MILENAGE */
-
 static const char *default_socket_path = "/tmp/hlr_auc_gw.sock";
 static const char *socket_path;
 static const char *default_gsm_triplet_file = "hostapd.sim_db";
 static const char *gsm_triplet_file;
 static int serv_sock = -1;
 
-/* OP and AMF parameters for Milenage (Example algorithms for AKA).
- * These example values are from 3GPP TS 35.208 v6.0.0 - 4.3.20 Test Set 20.
- */
-static const unsigned char milenage_op[16] = {
-	0x3f, 0xfc, 0xfe, 0x5b, 0x7b, 0x11, 0x11, 0x58,
-	0x99, 0x20, 0xd3, 0x52, 0x8e, 0x84, 0xe6, 0x55
+/* OPc and AMF parameters for Milenage (Example algorithms for AKA). */
+struct milenage_parameters {
+	struct milenage_parameters *next;
+	char imsi[20];
+	u8 ki[16];
+	u8 opc[16];
+	u8 amf[2];
+	u8 sqn[6];
 };
-static const unsigned char milenage_amf[2] = { 0x61, 0xdf };
-
-
-/* TODO: Add support for per-IMSI K and SQN from a text file. */
-
-/* Test K value from 3GPP TS 35.208 v6.0.0 - 4.3.20 Test Set 20. */
-static const u8 test_k[16] = {
-	0x90, 0xdc, 0xa4, 0xed, 0xa4, 0x5b, 0x53, 0xcf,
-	0x0f, 0x12, 0xd7, 0xc9, 0xc3, 0xbc, 0x6a, 0x89
-};
-static u8 test_sqn[6] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
 
+static struct milenage_parameters *milenage_db = NULL;
 
 #define EAP_SIM_MAX_CHAL 3
 
@@ -119,6 +107,145 @@ static int open_socket(const char *path)
 }
 
 
+static int read_milenage(const char *fname)
+{
+	FILE *f;
+	char buf[200], *pos, *pos2;
+	struct milenage_parameters *m = NULL;
+	int line, ret = 0;
+
+	if (fname == NULL)
+		return -1;
+
+	f = fopen(fname, "r");
+	if (f == NULL) {
+		printf("Could not open Milenage data file '%s'\n", fname);
+		return -1;
+	}
+
+	line = 0;
+	while (fgets(buf, sizeof(buf), f)) {
+		line++;
+
+		/* Parse IMSI Ki OPc AMF SQN */
+		buf[sizeof(buf) - 1] = '\0';
+		if (buf[0] == '#')
+			continue;
+		pos = buf;
+		while (*pos != '\0' && *pos != '\n')
+			pos++;
+		if (*pos == '\n')
+			*pos = '\0';
+		pos = buf;
+		if (*pos == '\0')
+			continue;
+
+		m = os_zalloc(sizeof(*m));
+		if (m == NULL) {
+			ret = -1;
+			break;
+		}
+
+		/* IMSI */
+		pos2 = strchr(pos, ' ');
+		if (pos2 == NULL) {
+			printf("%s:%d - Invalid IMSI (%s)\n",
+			       fname, line, pos);
+			ret = -1;
+			break;
+		}
+		*pos2 = '\0';
+		if (strlen(pos) >= sizeof(m->imsi)) {
+			printf("%s:%d - Too long IMSI (%s)\n",
+			       fname, line, pos);
+			ret = -1;
+			break;
+		}
+		strncpy(m->imsi, pos, sizeof(m->imsi));
+		pos = pos2 + 1;
+
+		/* Ki */
+		pos2 = strchr(pos, ' ');
+		if (pos2 == NULL) {
+			printf("%s:%d - Invalid Ki (%s)\n", fname, line, pos);
+			ret = -1;
+			break;
+		}
+		*pos2 = '\0';
+		if (strlen(pos) != 32 || hexstr2bin(pos, m->ki, 16)) {
+			printf("%s:%d - Invalid Ki (%s)\n", fname, line, pos);
+			ret = -1;
+			break;
+		}
+		pos = pos2 + 1;
+
+		/* OPc */
+		pos2 = strchr(pos, ' ');
+		if (pos2 == NULL) {
+			printf("%s:%d - Invalid OPc (%s)\n", fname, line, pos);
+			ret = -1;
+			break;
+		}
+		*pos2 = '\0';
+		if (strlen(pos) != 32 || hexstr2bin(pos, m->opc, 16)) {
+			printf("%s:%d - Invalid OPc (%s)\n", fname, line, pos);
+			ret = -1;
+			break;
+		}
+		pos = pos2 + 1;
+
+		/* AMF */
+		pos2 = strchr(pos, ' ');
+		if (pos2 == NULL) {
+			printf("%s:%d - Invalid AMF (%s)\n", fname, line, pos);
+			ret = -1;
+			break;
+		}
+		*pos2 = '\0';
+		if (strlen(pos) != 4 || hexstr2bin(pos, m->amf, 2)) {
+			printf("%s:%d - Invalid AMF (%s)\n", fname, line, pos);
+			ret = -1;
+			break;
+		}
+		pos = pos2 + 1;
+
+		/* SQN */
+		pos2 = strchr(pos, ' ');
+		if (pos2)
+			*pos2 = '\0';
+		if (strlen(pos) != 12 || hexstr2bin(pos, m->sqn, 6)) {
+			printf("%s:%d - Invalid SEQ (%s)\n", fname, line, pos);
+			ret = -1;
+			break;
+		}
+		pos = pos2 + 1;
+
+		m->next = milenage_db;
+		milenage_db = m;
+		m = NULL;
+	}
+	free(m);
+
+	fclose(f);
+
+	return ret;
+}
+
+
+static struct milenage_parameters * get_milenage(const char *imsi)
+{
+	struct milenage_parameters *m = milenage_db;
+
+	while (m) {
+		if (strcmp(m->imsi, imsi) == 0)
+			break;
+		m = m->next;
+	}
+
+	return m;
+}
+
+
 static void sim_req_auth(int s, struct sockaddr_un *from, socklen_t fromlen,
 			 char *imsi)
 {
@@ -126,6 +253,7 @@ static void sim_req_auth(int s, struct sockaddr_un *from, socklen_t fromlen,
 	int count, max_chal, ret;
 	char buf[80], *pos;
 	char reply[1000], *rpos, *rend;
+	struct milenage_parameters *m;
 
 	reply[0] = '\0';
 
@@ -145,6 +273,23 @@ static void sim_req_auth(int s, struct sockaddr_un *from, socklen_t fromlen,
 		return;
 	rpos += ret;
 
+	m = get_milenage(imsi);
+	if (m) {
+		u8 _rand[16], sres[4], kc[8];
+		for (count = 0; count < max_chal; count++) {
+			os_get_random(_rand, 16);
+			gsm_milenage(m->opc, m->ki, _rand, sres, kc);
+			*rpos++ = ' ';
+			rpos += wpa_snprintf_hex(rpos, rend - rpos, kc, 8);
+			*rpos++ = ':';
+			rpos += wpa_snprintf_hex(rpos, rend - rpos, sres, 4);
+			*rpos++ = ':';
+			rpos += wpa_snprintf_hex(rpos, rend - rpos, _rand, 16);
+		}
+		*rpos = '\0';
+		goto send;
+	}
+
 	/* TODO: could read triplet file into memory during startup and then
 	 * have pointer for IMSI to allow more than three first entries to be
 	 * used. */
@@ -179,8 +324,10 @@ static void sim_req_auth(int s, struct sockaddr_un *from, socklen_t fromlen,
 			continue;
 
 		ret = snprintf(rpos, rend - rpos, " %s", pos);
-		if (ret < 0 || ret >= rend - rpos)
+		if (ret < 0 || ret >= rend - rpos) {
+			fclose(f);
 			return;
+		}
 		rpos += ret;
 		count++;
 	}
@@ -208,31 +355,39 @@ static void aka_req_auth(int s, struct sockaddr_un *from, socklen_t fromlen,
 {
 	/* AKA-RESP-AUTH <IMSI> <RAND> <AUTN> <IK> <CK> <RES> */
 	char reply[1000], *pos, *end;
-	u8 rand[EAP_AKA_RAND_LEN];
+	u8 _rand[EAP_AKA_RAND_LEN];
 	u8 autn[EAP_AKA_AUTN_LEN];
 	u8 ik[EAP_AKA_IK_LEN];
 	u8 ck[EAP_AKA_CK_LEN];
 	u8 res[EAP_AKA_RES_MAX_LEN];
 	size_t res_len;
 	int ret;
+	struct milenage_parameters *m;
 
-#ifdef AKA_USE_MILENAGE
-	os_get_random(rand, EAP_AKA_RAND_LEN);
-	res_len = EAP_AKA_RES_MAX_LEN;
-	inc_byte_array(test_sqn, 6);
-	printf("AKA: Milenage with SQN=%02x%02x%02x%02x%02x%02x\n",
-	       test_sqn[0], test_sqn[1], test_sqn[2],
-	       test_sqn[3], test_sqn[4], test_sqn[5]);
-	milenage_generate(milenage_op, milenage_amf, test_k, test_sqn, rand,
-			  autn, ik, ck, res, &res_len);
-#else
-	memset(rand, '0', EAP_AKA_RAND_LEN);
-	memset(autn, '1', EAP_AKA_AUTN_LEN);
-	memset(ik, '3', EAP_AKA_IK_LEN);
-	memset(ck, '4', EAP_AKA_CK_LEN);
-	memset(res, '2', EAP_AKA_RES_MAX_LEN);
-	res_len = EAP_AKA_RES_MAX_LEN;
-#endif
+	m = get_milenage(imsi);
+	if (m) {
+		os_get_random(_rand, EAP_AKA_RAND_LEN);
+		res_len = EAP_AKA_RES_MAX_LEN;
+		inc_byte_array(m->sqn, 6);
+		printf("AKA: Milenage with SQN=%02x%02x%02x%02x%02x%02x\n",
+		       m->sqn[0], m->sqn[1], m->sqn[2],
+		       m->sqn[3], m->sqn[4], m->sqn[5]);
+		milenage_generate(m->opc, m->amf, m->ki, m->sqn, _rand,
+				  autn, ik, ck, res, &res_len);
+	} else {
+		printf("Unknown IMSI: %s\n", imsi);
+#ifdef AKA_USE_FIXED_TEST_VALUES
+		printf("Using fixed test values for AKA\n");
+		memset(_rand, '0', EAP_AKA_RAND_LEN);
+		memset(autn, '1', EAP_AKA_AUTN_LEN);
+		memset(ik, '3', EAP_AKA_IK_LEN);
+		memset(ck, '4', EAP_AKA_CK_LEN);
+		memset(res, '2', EAP_AKA_RES_MAX_LEN);
+		res_len = EAP_AKA_RES_MAX_LEN;
+#else /* AKA_USE_FIXED_TEST_VALUES */
+		return;
+#endif /* AKA_USE_FIXED_TEST_VALUES */
+	}
 
 	pos = reply;
 	end = &reply[sizeof(reply)];
@@ -240,7 +395,7 @@ static void aka_req_auth(int s, struct sockaddr_un *from, socklen_t fromlen,
 	if (ret < 0 || ret >= end - pos)
 		return;
 	pos += ret;
-	pos += wpa_snprintf_hex(pos, end - pos, rand, EAP_AKA_RAND_LEN);
+	pos += wpa_snprintf_hex(pos, end - pos, _rand, EAP_AKA_RAND_LEN);
 	*pos++ = ' ';
 	pos += wpa_snprintf_hex(pos, end - pos, autn, EAP_AKA_AUTN_LEN);
 	*pos++ = ' ';
@@ -249,7 +404,6 @@ static void aka_req_auth(int s, struct sockaddr_un *from, socklen_t fromlen,
 	pos += wpa_snprintf_hex(pos, end - pos, ck, EAP_AKA_CK_LEN);
 	*pos++ = ' ';
 	pos += wpa_snprintf_hex(pos, end - pos, res, res_len);
-	*pos++ = ' ';
 
 	printf("Send: %s\n", reply);
 
@@ -264,6 +418,7 @@ static void aka_auts(int s, struct sockaddr_un *from, socklen_t fromlen,
 {
 	char *auts, *rand;
 	u8 _auts[EAP_AKA_AUTS_LEN], _rand[EAP_AKA_RAND_LEN], sqn[6];
+	struct milenage_parameters *m;
 
 	/* AKA-AUTS <IMSI> <AUTS> <RAND> */
 
@@ -284,10 +439,16 @@ static void aka_auts(int s, struct sockaddr_un *from, socklen_t fromlen,
 		return;
 	}
 
-	if (milenage_auts(milenage_op, test_k, _rand, _auts, sqn)) {
+	m = get_milenage(imsi);
+	if (m == NULL) {
+		printf("Unknown IMSI: %s\n", imsi);
+		return;
+	}
+
+	if (milenage_auts(m->opc, m->ki, _rand, _auts, sqn)) {
 		printf("AKA-AUTS: Incorrect MAC-S\n");
 	} else {
-		memcpy(test_sqn, sqn, 6);
+		memcpy(m->sqn, sqn, 6);
 		printf("AKA-AUTS: Re-synchronized: "
 		       "SQN=%02x%02x%02x%02x%02x%02x\n",
 		       sqn[0], sqn[1], sqn[2], sqn[3], sqn[4], sqn[5]);
@@ -334,6 +495,15 @@ static int process(int s)
 
 static void cleanup(void)
 {
+	struct milenage_parameters *m, *prev;
+
+	m = milenage_db;
+	while (m) {
+		prev = m;
+		m = m->next;
+		free(prev);
+	}
+
 	close(serv_sock);
 	unlink(socket_path);
 }
@@ -353,14 +523,16 @@ static void usage(void)
 	       "Copyright (c) 2005-2006, Jouni Malinen <jkmaline@cc.hut.fi>\n"
 	       "\n"
 	       "usage:\n"
-	       "hlr_auc_gw [-h] [-s<socket path>] [-g<triplet file>]\n"
+	       "hlr_auc_gw [-h] [-s<socket path>] [-g<triplet file>] "
+	       "[-m<milenage file>]\n"
 	       "\n"
 	       "options:\n"
 	       "  -h = show this usage help\n"
 	       "  -s<socket path> = path for UNIX domain socket\n"
 	       "                    (default: %s)\n"
 	       "  -g<triplet file> = path for GSM authentication triplets\n"
-	       "                     (default: %s)\n",
+	       "                     (default: %s)\n"
+	       "  -m<milenage file> = path for Milenage keys\n",
 	       default_socket_path, default_gsm_triplet_file);
 }
 
@@ -368,12 +540,13 @@ static void usage(void)
 int main(int argc, char *argv[])
 {
 	int c;
+	char *milenage_file = NULL;
 
 	socket_path = default_socket_path;
 	gsm_triplet_file = default_gsm_triplet_file;
 
 	for (;;) {
-		c = getopt(argc, argv, "g:hs:");
+		c = getopt(argc, argv, "g:hm:s:");
 		if (c < 0)
 			break;
 		switch (c) {
@@ -383,6 +556,9 @@ int main(int argc, char *argv[])
 		case 'h':
 			usage();
 			return 0;
+		case 'm':
+			milenage_file = optarg;
+			break;
 		case 's':
 			socket_path = optarg;
 			break;
@@ -392,6 +568,9 @@ int main(int argc, char *argv[])
 		}
 	}
 
+	if (milenage_file && read_milenage(milenage_file) < 0)
+		return -1;
+
 	serv_sock = open_socket(socket_path);
 	if (serv_sock < 0)
 		return -1;
