@@ -166,9 +166,105 @@ static int hostapd_prepare_rates(struct hostapd_data *hapd,
 }
 
 
-int hostapd_select_hw_mode(struct hostapd_iface *iface)
+static void select_hw_mode_start(void *eloop_data, void *user_ctx);
+static void select_hw_mode2_handler(void *eloop_data, void *user_ctx);
+
+/**
+ * select_hw_mode_finalize - Finish select HW mode & call the callback
+ * @iface: Pointer to interface data.
+ * @status: Status of the select HW mode (0 on success; -1 on failure).
+ * Returns: 0 on success; -1 on failure (e.g., was not in progress).
+ */
+static int select_hw_mode_finalize(struct hostapd_iface *iface, int status)
 {
-	int i, j, ok, res;
+	hostapd_iface_cb cb;
+
+	if (!iface->hw_mode_sel_cb)
+		return -1;
+
+	eloop_cancel_timeout(select_hw_mode_start, iface, NULL);
+	eloop_cancel_timeout(select_hw_mode2_handler, iface, NULL);
+
+	cb = iface->hw_mode_sel_cb;
+
+	iface->hw_mode_sel_cb = NULL;
+
+	cb(iface, status);
+
+	return 0;
+}
+
+
+/**
+ * select_hw_mode2 - Select the hardware mode (part 2)
+ * @iface: Pointer to interface data.
+ * @status: Status of auto chanel selection.
+ *
+ * Setup the rates and passive scanning based on the configuration.
+ */
+static void select_hw_mode2(struct hostapd_iface *iface, int status)
+{
+	int ret = status;
+	if (ret)
+		goto fail;
+
+	if (iface->current_mode == NULL) {
+		hostapd_logger(iface->bss[0], NULL, HOSTAPD_MODULE_IEEE80211,
+			       HOSTAPD_LEVEL_WARNING,
+			       "Hardware does not support configured channel");
+		ret = -1;
+		goto fail;
+	}
+
+	if (hostapd_prepare_rates(iface->bss[0], iface->current_mode)) {
+		printf("Failed to prepare rates table.\n");
+		hostapd_logger(iface->bss[0], NULL, HOSTAPD_MODULE_IEEE80211,
+					   HOSTAPD_LEVEL_WARNING,
+					   "Failed to prepare rates table.");
+		ret = -1;
+		goto fail;
+	}
+
+	ret = hostapd_passive_scan(iface->bss[0], 0,
+				   iface->conf->passive_scan_mode,
+				   iface->conf->passive_scan_interval,
+				   iface->conf->passive_scan_listen,
+				   NULL, NULL);
+	if (ret) {
+		printf("Could not set passive scanning: %s\n", strerror(ret));
+		ret = 0;
+	}
+
+fail:
+	select_hw_mode_finalize(iface, ret);
+}
+
+
+/**
+ * select_hw_mode2_handler - Calls select_hw_mode2 when auto chan isn't used
+ * @eloop_data: Stores the struct hostapd_iface * for the interface.
+ * @user_ctx: Unused.
+ */
+static void select_hw_mode2_handler(void *eloop_data, void *user_ctx)
+{
+	struct hostapd_iface *iface = eloop_data;
+
+	select_hw_mode2(iface, 0);
+}
+
+
+/**
+ * select_hw_mode1 - Select the hardware mode (part 1)
+ * @iface: Pointer to interface data.
+ * Returns: 0 on success; -1 on failure.
+ *
+ * Setup the hardware mode and channel based on the configuration.
+ * Schedules select_hw_mode2() to be called immediately or after automatic
+ * channel selection takes place.
+ */
+static int select_hw_mode1(struct hostapd_iface *iface)
+{
+	int i, j, ok;
 
 	if (iface->num_hw_features < 1)
 		return -1;
@@ -186,7 +282,8 @@ int hostapd_select_hw_mode(struct hostapd_iface *iface)
 		printf("Hardware does not support configured mode\n");
 		hostapd_logger(iface->bss[0], NULL, HOSTAPD_MODULE_IEEE80211,
 			       HOSTAPD_LEVEL_WARNING,
-			       "Hardware does not support configured mode");
+			       "Hardware does not support configured mode "
+			       "(%d)", (int) iface->conf->hw_mode);
 		return -1;
 	}
 
@@ -212,35 +309,75 @@ int hostapd_select_hw_mode(struct hostapd_iface *iface)
 		iface->current_mode = NULL;
 	}
 
-	if (iface->current_mode == NULL) {
-		hostapd_logger(iface->bss[0], NULL, HOSTAPD_MODULE_IEEE80211,
-			       HOSTAPD_LEVEL_WARNING,
-			       "Hardware does not support configured channel");
+	/*
+	 * Calls select_hw_mode2() via a handler, so that the function is
+	 * always executed from eloop.
+	 */
+	eloop_register_timeout(0, 0, select_hw_mode2_handler, iface, NULL);
+	return 0;
+}
+
+
+/**
+ * select_hw_mode_start - Handler to start select HW mode
+ * @eloop_data: Stores the struct hostapd_iface * for the interface.
+ * @user_ctx: Unused.
+ *
+ * An eloop handler is used so that all errors can be processed by the
+ * callback without introducing stack recursion.
+ */
+static void select_hw_mode_start(void *eloop_data, void *user_ctx)
+{
+	struct hostapd_iface *iface = (struct hostapd_iface *)eloop_data;
+
+	int ret;
+
+	ret = select_hw_mode1(iface);
+	if (ret)
+		select_hw_mode_finalize(iface, ret);
+}
+
+
+/**
+ * hostapd_select_hw_mode_start - Start selection of the hardware mode
+ * @iface: Pointer to interface data.
+ * @cb: The function to callback when done.
+ * Returns:  0 if it starts successfully; cb will be called when done.
+ *          -1 on failure; cb will not be called.
+ *
+ * Sets up the hardware mode, channel, rates, and passive scanning
+ * based on the configuration.
+ */
+int hostapd_select_hw_mode_start(struct hostapd_iface *iface,
+				 hostapd_iface_cb cb)
+{
+	if (iface->hw_mode_sel_cb) {
+		wpa_printf(MSG_DEBUG,
+			   "%s: Hardware mode select already in progress.",
+			   iface->bss[0]->conf->iface);
 		return -1;
 	}
 
-	if (hostapd_prepare_rates(iface->bss[0], iface->current_mode)) {
-		printf("Failed to prepare rates table.\n");
-		hostapd_logger(iface->bss[0], NULL, HOSTAPD_MODULE_IEEE80211,
-					   HOSTAPD_LEVEL_WARNING,
-					   "Failed to prepare rates table.");
-		return -1;
-	}
+	iface->hw_mode_sel_cb = cb;
 
-	res = hostapd_passive_scan(iface->bss[0], 0,
-				   iface->conf->passive_scan_mode,
-				   iface->conf->passive_scan_interval,
-				   iface->conf->passive_scan_listen,
-				   NULL, NULL);
-	if (res) {
-		printf("Could not set passive scanning: %s\n",
-		       strerror(res));
-	}
+	eloop_register_timeout(0, 0, select_hw_mode_start, iface, NULL);
 
 	return 0;
 }
 
 
+/**
+ * hostapd_auto_chan_select_stop - Stops automatic channel selection
+ * @iface: Pointer to interface data.
+ * Returns:  0 if successfully stopped;
+ *          -1 on failure (i.e., was not in progress)
+ */
+int hostapd_select_hw_mode_stop(struct hostapd_iface *iface)
+{
+	return select_hw_mode_finalize(iface, -1);
+}
+
+
 const char * hostapd_hw_mode_txt(int mode)
 {
 	switch (mode) {
